@@ -26,6 +26,8 @@ copy_range_help(void)
 					       file at offset 200\n\
  'copy_range some_file' - copies all bytes from some_file into the open file\n\
                           at position 0\n\
+ 'copy_range -f 2' - copies all bytes from open file 2 into the current open file\n\
+                          at position 0\n\
 "));
 }
 
@@ -34,13 +36,13 @@ copy_range_help(void)
  * glibc buffered copy fallback.
  */
 static loff_t
-copy_file_range_cmd(int fd, long long *src, long long *dst, long long len)
+copy_file_range_cmd(int fd, long long *src_off, long long *dst_off, size_t len)
 {
 	loff_t ret;
 
 	do {
-		ret = syscall(__NR_copy_file_range, fd, src, file->fd, dst,
-				len, 0);
+		ret = syscall(__NR_copy_file_range, fd, src_off,
+				file->fd, dst_off, len, 0);
 		if (ret == -1) {
 			perror("copy_range");
 			return errno;
@@ -64,67 +66,84 @@ copy_src_filesize(int fd)
 	return st.st_size;
 }
 
-static int
-copy_dst_truncate(void)
-{
-	int ret = ftruncate(file->fd, 0);
-	if (ret < 0)
-		perror("ftruncate");
-	return ret;
-}
-
 static int
 copy_range_f(int argc, char **argv)
 {
-	long long src = 0;
-	long long dst = 0;
-	long long len = 0;
+	long long src_off = 0;
+	long long dst_off = 0;
+	size_t len = 0;
+	bool len_specified = false;
 	int opt;
 	int ret;
 	int fd;
+	int src_path_arg = 1;
+	int src_file_nr = 0;
 	size_t fsblocksize, fssectsize;
 
 	init_cvtnum(&fsblocksize, &fssectsize);
 
-	while ((opt = getopt(argc, argv, "s:d:l:")) != -1) {
+	while ((opt = getopt(argc, argv, "s:d:l:f:")) != -1) {
 		switch (opt) {
 		case 's':
-			src = cvtnum(fsblocksize, fssectsize, optarg);
-			if (src < 0) {
+			src_off = cvtnum(fsblocksize, fssectsize, optarg);
+			if (src_off < 0) {
 				printf(_("invalid source offset -- %s\n"), optarg);
 				return 0;
 			}
 			break;
 		case 'd':
-			dst = cvtnum(fsblocksize, fssectsize, optarg);
-			if (dst < 0) {
+			dst_off = cvtnum(fsblocksize, fssectsize, optarg);
+			if (dst_off < 0) {
 				printf(_("invalid destination offset -- %s\n"), optarg);
 				return 0;
 			}
 			break;
 		case 'l':
 			len = cvtnum(fsblocksize, fssectsize, optarg);
-			if (len < 0) {
+			if (len == -1LL) {
 				printf(_("invalid length -- %s\n"), optarg);
 				return 0;
 			}
+			len_specified = true;
+			break;
+		case 'f':
+			src_file_nr = atoi(argv[1]);
+			if (src_file_nr < 0 || src_file_nr >= filecount) {
+				printf(_("file value %d is out of range (0-%d)\n"),
+					src_file_nr, filecount - 1);
+				return 0;
+			}
+			/* Expect no src_path arg */
+			src_path_arg = 0;
 			break;
 		}
 	}
 
-	if (optind != argc - 1)
+	if (optind != argc - src_path_arg)
 		return command_usage(&copy_range_cmd);
 
-	fd = openfile(argv[optind], NULL, IO_READONLY, 0, NULL);
-	if (fd < 0)
-		return 0;
-
-	if (src == 0 && dst == 0 && len == 0) {
-		len = copy_src_filesize(fd);
-		copy_dst_truncate();
+	if (src_path_arg) {
+		fd = openfile(argv[optind], NULL, IO_READONLY, 0, NULL);
+		if (fd < 0)
+			return 0;
+	} else {
+		fd = filetable[src_file_nr].fd;
 	}
 
-	ret = copy_file_range_cmd(fd, &src, &dst, len);
+	if (!len_specified) {
+		off64_t	sz;
+
+		sz = copy_src_filesize(fd);
+		if (sz < 0 || (unsigned long long)sz > SIZE_MAX) {
+			ret = 1;
+			goto out;
+		}
+		if (sz > src_off)
+			len = sz - src_off;
+	}
+
+	ret = copy_file_range_cmd(fd, &src_off, &dst_off, len);
+out:
 	close(fd);
 	return ret;
 }
@@ -137,7 +156,7 @@ copy_range_init(void)
 	copy_range_cmd.argmin = 1;
 	copy_range_cmd.argmax = 7;
 	copy_range_cmd.flags = CMD_NOMAP_OK | CMD_FOREIGN_OK;
-	copy_range_cmd.args = _("[-s src_off] [-d dst_off] [-l len] src_file");
+	copy_range_cmd.args = _("[-s src_off] [-d dst_off] [-l len] src_file | -f N");
 	copy_range_cmd.oneline = _("Copy a range of data between two files");
 	copy_range_cmd.help = copy_range_help;
 
